module huber_bfgs_b_dp
 
  use cp_lbfgs_dp
  implicit none
  
!	INTEGER, PARAMETER      :: dbl = SELECTED_REAL_KIND(15,307)
contains
  
  subroutine lbfgs(oper, reg, fonction, gradient, xc, niter, accuracy, &
  & boundtype, low, up, verb, x0, err,its) 
    optional               :: x0, err,its
    interface 

       function oper( adj, add, x, dat) result(stat)
         integer             :: stat
         logical, intent(in) :: adj, add
         real(kind=8), dimension(: ) :: x, dat
       end function oper

       function reg( adj, add, x, dat) result(stat)
         integer             :: stat
         logical, intent(in) :: adj, add
         real(kind=8), dimension(:) :: x, dat
       end function reg

       function fonction(oper,reg,x,fct) result (stat)

         interface
            function oper( adj, add, x, dat) result(stat)
    integer             :: stat
              logical, intent(in) :: adj, add
              real(kind=8), dimension(:) :: x, dat
            end function oper
            function reg( adj, add, x, dat) result(stat)
              integer             :: stat
              logical, intent(in) :: adj, add
              real(kind=8), dimension(:) :: x, dat
            end function reg
         end interface

         integer             :: stat
         real(kind=8),dimension(:)   :: x
         real(kind=8)                :: fct

       end function fonction

       function gradient(oper,reg,x,g) result (stat)

         interface
            function oper( adj, add, x, dat) result(stat)
	      integer             :: stat
              logical, intent(in) :: adj, add
              real(kind=8), dimension(: ) :: x, dat
            end function oper
            function reg( adj, add, x, dat) result(stat)
              integer             :: stat
              logical, intent(in) :: adj, add
              real(kind=8), dimension(: ) :: x, dat
            end function reg
         end interface

         integer             :: stat
         real(kind=8),dimension(:)   :: x,g

       end function gradient


    end interface


    real(kind=8), dimension(:),  intent( in)    :: x0                ! initial guess
    integer,                       intent( in)    :: niter, boundtype  ! iterations
    real(kind=8), dimension(:),  intent(out)    :: xc,err            ! solution and error vector
    real(kind=8), dimension(:),  allocatable    :: g                 ! gradient
    integer                                       :: iter,stat1,stat2,stat3,its
    real(kind=8)                                :: res, f, f0
    real(kind=8)                                :: alpha             ! step after line search
    integer                                       :: info, verb

    ! L-BFGS-B variables
    real(kind=8)                            :: dsave(29),factr,pgtol
    real(kind=8),intent(in)                 :: accuracy
    real(kind=8),intent(in)                 :: low(:), up(:)
    integer                         :: mmax, nmax
    integer,dimension(:),allocatable:: iwa, nbd
    real(kind=8),                allocatable:: wa(:)

    character*60                    :: task, csave
    logical                         :: lsave(4), cont
    integer                         :: isave(44), mmem, iprint

    xc=0.
    if (present (x0))  then
       xc=x0
    end if

    nmax = size(xc)
    mmax = 5                                        

    allocate(g(size(xc)))
    allocate(wa(2*mmax*nmax+4*nmax+12*mmax*mmax+12*mmax))
    allocate(iwa(3*nmax))
    allocate(nbd(size(xc)))

    nbd = boundtype
    iprint=0
    mmem=5
    factr=1.0e+1
    pgtol=accuracy
    cont=.true.; task = 'START';iter=0

    stat1 = fonction(oper,reg,xc,f)
    stat2 = gradient(oper,reg,xc,g)
    f0=10000.
    do while (cont.and.(iter.le.niter))

       cont=.false.

       call setulb(size(xc),mmem,xc,low,up,nbd,f,g,factr,pgtol,wa,iwa,task,iprint,&
       & csave,lsave,isave,dsave)

       if (task(1:2) .eq. 'FG') then

          stat1 = fonction(oper,reg,xc,f)
          stat2 = gradient(oper,reg,xc,g)

          cont=.true.

       end if

       if (task(1:5) .eq. 'NEW_X') then

          cont=.true.
          iter = iter+1

          if (verb.gt.0) then
             write(0,*) 'iter=',iter,'fct=',f,'diff=',f0-f
             f0=f 
         end if
 
       end if

    end do

    deallocate(g,nbd,wa,iwa)
    its=iter
  end subroutine lbfgs
  
end module huber_bfgs_b_dp
